/*
 *      Copyright (c) 2018-2028, Chill Zhuang All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *  Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in the
 *  documentation and/or other materials provided with the distribution.
 *  Neither the name of the dreamlu.net developer nor the names of its
 *  contributors may be used to endorse or promote products derived from
 *  this software without specific prior written permission.
 *  Author: Chill 庄骞 (smallchill@163.com)
 */
package org.springblade.csetl.controller;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.github.xiaoymin.knife4j.annotations.ApiOperationSupport;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import io.swagger.annotations.ApiParam;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.springblade.core.boot.ctrl.BladeController;
import org.springblade.core.mp.support.Condition;
import org.springblade.core.mp.support.Query;
import org.springblade.core.tool.api.R;
import org.springblade.core.tool.utils.Func;
import org.springblade.csetl.entity.CsReportDatasource;
import org.springblade.csetl.service.CsReportDatasourceService;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.util.List;

/**
 * 报表数据源 控制器
 *
 * @author BladeX
 * @since 2021-03-22
 */
@Slf4j
@RestController
@RequestMapping("/cs/report/reportdatasource")
@Api(value = "数据源", tags = "数据源")
public class CsReportDatasourceController extends BladeController {

    @Autowired
    private CsReportDatasourceService reportDatasourceService;


    private final static String DEFAULT_ENCRYPT_PASSWORD = "******";

    /**
     * 详情
     */
    @GetMapping("/detail")
    @ApiOperationSupport(order = 1)
    @ApiOperation(value = "详情 detail", notes = "传入reportDatasource")
    public R<CsReportDatasource> detail(CsReportDatasource reportDatasource) {
        CsReportDatasource detail = reportDatasourceService.getOne(Condition.getQueryWrapper(reportDatasource));
        return R.data(detail);
    }

    /**
     * 分页 报表数据源
     */
    @GetMapping("/list")
    @ApiOperationSupport(order = 2)
    @ApiOperation(value = "列表 list", notes = "传入reportDatasource")
    @Deprecated
    public R<List<CsReportDatasource>> list(CsReportDatasource reportDatasource) {
        LambdaQueryWrapper<CsReportDatasource> queryWrapper = Condition.getQueryWrapper(reportDatasource).lambda().orderByDesc(CsReportDatasource::getId);
        List<CsReportDatasource> dataList = reportDatasourceService.list(queryWrapper);
        if (CollectionUtils.isNotEmpty(dataList)) {
            dataList.forEach(item -> item.setDsPassword(DEFAULT_ENCRYPT_PASSWORD));
        }
        return R.data(dataList);
    }

    /**
     * 自定义分页 报表数据源
     */
    @GetMapping("/page")
    @ApiOperationSupport(order = 3)
    @ApiOperation(value = "分页 page", notes = "传入reportDatasource")
    public R<IPage<CsReportDatasource>> page(CsReportDatasource reportDatasource, Query query) {
        LambdaQueryWrapper<CsReportDatasource> queryWrapper = Condition.getQueryWrapper(reportDatasource).lambda().orderByDesc(CsReportDatasource::getId);
        IPage<CsReportDatasource> page = reportDatasourceService.page(Condition.getPage(query), queryWrapper);
        if (page != null && CollectionUtils.isNotEmpty(page.getRecords())) {
            page.getRecords().forEach(item -> item.setDsPassword(DEFAULT_ENCRYPT_PASSWORD));
        }
        return R.data(page);
    }

    /**
     * 新增 报表数据源
     */
    @PostMapping("/save")
    @ApiOperationSupport(order = 4)
    @ApiOperation(value = "新增 save", notes = "传入reportDatasource")
    public R<?> save(@Valid @RequestBody CsReportDatasource dto) {
        dto.setCreateUserId(getUser().getUserId());
        //ETL默认不添加数据源到连接池【使用懒加载模式】
        return R.status(
                reportDatasourceService.testDs(dto) && reportDatasourceService.save(dto));
    }

    /**
     * 修改 报表数据源
     */
    @PostMapping("/update")
    @ApiOperationSupport(order = 5)
    @ApiOperation(value = "修改 update", notes = "传入reportDatasource")
    public R<?> update(@Valid @RequestBody CsReportDatasource dto) {
        if (dto.getId() == null || dto.getId() <= 0) {
            return R.fail("数据源ID不能为空");
        }
        CsReportDatasource dbReportDs = reportDatasourceService.getById(dto.getId());
        if (dbReportDs == null) {
            return R.fail("数据源不存在");
        }
        try {
            if (DEFAULT_ENCRYPT_PASSWORD.equals(dto.getDsPassword())) {
                dto.setDsPassword(dbReportDs.getDsPassword());
            }
            //ETL默认不添加数据源到连接池【使用懒加载模式】
//			boolean ret = dynamicDataSourceSupport.addDs(dto);
            return R.status(reportDatasourceService.testDs(dto) && reportDatasourceService.updateById(dto));
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            return R.fail("修改数据源失败");
        }
    }


    /**
     * 删除 报表数据源
     */
    @PostMapping("/remove")
    @ApiOperationSupport(order = 6)
    @ApiOperation(value = "删除 remove", notes = "传入ids")
    public R<?> remove(@ApiParam(value = "主键集合", required = true) @RequestParam String ids) {
        List<Long> dsIds = Func.toLongList(ids);
        return R.status(reportDatasourceService.removeByIds(dsIds));
    }

    /**
     * 测试数据源
     *
     * @param dto
     * @return
     */
    @PostMapping("/testDs")
    @ApiOperationSupport(order = 7)
    @ApiOperation(value = "测试数据源 testDs", notes = "传入数据源信息")
    public R<?> testDs(@Validated @RequestBody CsReportDatasource dto) {
        //解决更新页面，密码是加密导致的测试数据源失败
        if (DEFAULT_ENCRYPT_PASSWORD.equals(dto.getDsPassword())) {
            CsReportDatasource dbReportDs = reportDatasourceService.getById(dto.getId());
            if (dbReportDs != null) {
                dto.setDsPassword(dbReportDs.getDsPassword());
            }
        }
        boolean testDs = reportDatasourceService.testDs(dto);
        return testDs ? R.success("连接成功") : R.fail("连接失败,请检查");
    }
}
