/*
 * Copyright (c) 2018-2028, AoKunSang All rights reserved.
 * 我问青山何时老，青山问我几时闲，
 * 不是闲人闲不得，能闲必非等闲人。
 */
package org.springblade.csetl.engine;

import cn.hutool.core.util.StrUtil;
import cn.hutool.json.JSONUtil;
import com.influxdb.client.InfluxDBClient;
import com.influxdb.client.InfluxDBClientFactory;
import com.influxdb.client.InfluxDBClientOptions;
import com.influxdb.query.FluxRecord;
import com.influxdb.query.FluxTable;
import lombok.extern.slf4j.Slf4j;
import org.influxdb.InfluxDB;
import org.influxdb.InfluxDBFactory;
import org.influxdb.dto.Query;
import org.influxdb.dto.QueryResult;
import org.springblade.core.tool.exception.ServiceException;
import org.springblade.csetl.dto.BaseQueryParam;
import org.springblade.csetl.dto.InfluxdbQueryParam;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * influxdb执行器
 *
 * @author wookvn
 * @datetime 2025-02-05 15:16:33
 */
@Slf4j
@Component
public class InfluxdbExecutor implements SqlExecutor<List<Map<String, Object>>> {

    @Override
    public Boolean check(BaseQueryParam queryParam) {
        InfluxdbQueryParam influxdbQueryParam = (InfluxdbQueryParam) queryParam;
        //v2.x版本
        if(influxdbQueryParam.getFluxV2Status()) {
            //如果有token，则不使用userpass
            try(InfluxDBClient client = InfluxDBClientFactory.create(_buildInfluxDBV2Options(influxdbQueryParam))) {
                return client.ping();
            } catch (Exception e) {
                log.error(e.getMessage(), e);
                return false;
            }
        }else {
            try(InfluxDB influxdb = InfluxDBFactory.connect(InfluxdbQueryParam.urlOf(influxdbQueryParam), influxdbQueryParam.getUsername(), influxdbQueryParam.getPassword());){
                return influxdb.ping().isGood();
            } catch (Exception e) {
                log.error(e.getMessage(), e);
                return false;
            }
        }
    }

    @Override
    public List<Map<String, Object>> query(BaseQueryParam queryParam) {
        InfluxdbQueryParam influxdbQueryParam = (InfluxdbQueryParam) queryParam;
        return influxdbQueryParam.getFluxV2Status() ? fluxQuery(influxdbQueryParam) : influxQLQuery(influxdbQueryParam);
    }

    /**
     * <p>influxdb-2.x查询方式</p>
     * @author wookvn
     * @datetime 2025-02-06 15:03:50
     * @param influxdbQueryParam: （参数说明）
     * @return java.util.List<java.util.Map<java.lang.String,java.lang.Object>>
     **/
    List<Map<String, Object>> fluxQuery(InfluxdbQueryParam influxdbQueryParam) {
        try(InfluxDBClient client = InfluxDBClientFactory.create(_buildInfluxDBV2Options(influxdbQueryParam))) {
            List<FluxTable> tables = client.getQueryApi().query(influxdbQueryParam.getQuerySql());
            List<Map<String, Object>> resultList = new ArrayList<>();
            for (FluxTable table : tables) {
                for (FluxRecord record : table.getRecords()) {
                    resultList.add(record.getValues());
                }
            }
            return resultList;
        }catch (Exception e){
            log.error(e.getMessage(), e);
        }
        return null;
    }

    /**
     * <p>influxdb-v1.x版本查询方式</p>
     * @author wookvn
     * @datetime 2025-02-06 15:04:10
     * @param influxdbQueryParam: （参数说明）
     * @return java.util.List<java.util.Map<java.lang.String,java.lang.Object>>
     **/
    List<Map<String, Object>> influxQLQuery(InfluxdbQueryParam influxdbQueryParam) {
        try(InfluxDB influxDB = InfluxDBFactory.connect(InfluxdbQueryParam.urlOf(influxdbQueryParam), influxdbQueryParam.getUsername(), influxdbQueryParam.getPassword());){
            QueryResult result = influxDB.query(new Query(influxdbQueryParam.getQuerySql(), influxdbQueryParam.getDatabase()));
            List<Map<String, Object>> resultList = new ArrayList<>();
            for (QueryResult.Result res : result.getResults()) {
                if(StrUtil.isNotBlank(res.getError())) {
                    throw new ServiceException("SQL执行错误：" + res.getError() + "，请检查SQL语句是否正确");
                }
                for (QueryResult.Series series : res.getSeries()) {
                    Map<String, Object> dataMap = new HashMap<>();
                    //字段名列表
                    List<String> columns = series.getColumns();
                    for (List<Object> row : series.getValues()) {
                        // 遍历字段（假设查询如 SELECT field1, field2 FROM ...）
                        for (int i = 0; i < row.size(); i++) {
                            String fieldName = columns.get(i);
                            Object value = row.get(i);
                            dataMap.put(fieldName, value);
                        }
                        resultList.add(dataMap);
                    }
                }
            }
            return resultList;
        }catch (Exception e){
            log.error(e.getMessage(), e);
        }
        return null;
    }

    private InfluxDBClientOptions _buildInfluxDBV2Options(InfluxdbQueryParam influxdbQueryParam) {
        InfluxDBClientOptions.Builder builder = InfluxDBClientOptions.builder().url(InfluxdbQueryParam.urlOf(influxdbQueryParam))
//                .bucket(influxdbQueryParam.getBucket())
                .org(influxdbQueryParam.getDatabase());
        if(StrUtil.isNotBlank(influxdbQueryParam.getToken())) {
            builder.authenticateToken(influxdbQueryParam.getToken().toCharArray());
        } else {
            builder.authenticate(influxdbQueryParam.getUsername(), influxdbQueryParam.getPassword().toCharArray());
        }
        return builder.build();
    }

    public static void main(String[] args) {
        InfluxdbExecutor influxdbExecutor = new InfluxdbExecutor();
        InfluxdbQueryParam influxdbQueryParam = new InfluxdbQueryParam();
        influxdbQueryParam.setHost("121.204.249.37");
        influxdbQueryParam.setPort(6086); //6086-v1.x版本 7086-v2.x版本
        influxdbQueryParam.setUsername("admin");
        influxdbQueryParam.setPassword("influxdb@123");
        influxdbQueryParam.setDatabase("platomix");
        influxdbQueryParam.setFluxV2Status(false);
        influxdbQueryParam.setToken("Dt40Jd8GJ15r9Cz-YoyQohGJMER0rJnmiKHHIhIuZFjRL7M615EawGLl_Jb6Wv7Onhx-Vle5hCUWgnzeGVLuoA==");
//V2的flux-sql
        //        influxdbQueryParam.setQuerySql("from(bucket: \"mybucket\")\n" +
//                "  |> range(start: -2h)\n" +
//                "  |> filter(fn: (r) => r[\"_measurement\"] == \"airSensors\" or r[\"_measurement\"] == \"average_temperature\" or r[\"_measurement\"] == \"h2o_feet\" or r[\"_measurement\"] == \"h2o_pH\" or r[\"_measurement\"] == \"h2o_quality\")\n" +
//                "  |> aggregateWindow(every: 1h, fn: mean, createEmpty: false)" +
//                ""
//        );
//V1的influxql
        influxdbQueryParam.setQuerySql("select * from \"treasures\"");
        System.out.println(influxdbQueryParam.getQuerySql());

        Boolean aBoolean = influxdbExecutor.check(influxdbQueryParam);
        System.out.println("check:" + aBoolean + "----- ------------");
        List<Map<String, Object>> mapList = influxdbExecutor.query(influxdbQueryParam);
        System.out.println("json:"+ JSONUtil.toJsonPrettyStr(mapList));
    }

}
