package org.springblade.etl.util;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.Properties;

import com.jcraft.jsch.Channel;
import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.SftpException;

import cn.hutool.core.io.IoUtil;
import lombok.extern.slf4j.Slf4j;

@Slf4j
public class SftpUtil {


    private ChannelSftp sftp;

    private Session session;
    /**
     * SFTP 登录用户名
     */
    private String username;
    /**
     * SFTP 登录密码
     */
    private String password;

    /**
     * SFTP 服务器地址IP地址
     */
    private String host;
    /**
     * SFTP 端口
     */
    private int port;


    /**
     * 构造基于密码认证的sftp对象
     */
    public SftpUtil(String username, String password, String host, int port) {
        this.username = username;
        this.password = password;
        this.host = host;
        this.port = port;
    }


    public SftpUtil() {
    }


    /**
     * 连接sftp服务器
     */
    public void login() throws Exception {
        JSch jsch = new JSch();

        session = jsch.getSession(username, host, port);

        if (password != null) {
            session.setPassword(password);
        }
        Properties config = new Properties();
        config.put("StrictHostKeyChecking", "no");
        //TODO 移除kerberos验证方法
        config.put("PreferredAuthentications", "publickey,keyboard-interactive,password");

        session.setConfig(config);
        session.setTimeout(20000);
        session.connect();

        Channel channel = session.openChannel("sftp");
        channel.connect();

        sftp = (ChannelSftp) channel;

    }

    /**
     * 关闭连接 server
     */
    public void logout() {
        if (sftp != null) {
            if (sftp.isConnected()) {
                sftp.disconnect();
            }
        }
        if (session != null) {
            if (session.isConnected()) {
                session.disconnect();
            }
        }
    }


    /**
     * 将输入流的数据上传到sftp作为文件。文件完整路径=basePath+directory
     *
     * @param path     上传到该目录
     * @param fileName sftp端文件名
     * @param input    输入流
     */
    public void upload(String path, String fileName, InputStream input) throws SftpException {
        try {
            sftp.cd(path);
        } catch (SftpException e) {
            //目录不存在，则创建文件夹
            String[] dirs = path.split("/");
            for (String dir : dirs) {
                if (null == dir || "".equals(dir)) {
                    continue;
                }
                try {
                    sftp.cd(dir);
                } catch (SftpException ex) {
                    sftp.mkdir(dir);
                    sftp.cd(dir);
                }
            }
        }
        //上传文件
        sftp.put(input, fileName);
    }


    /**
     * 读取文件
     */
    public InputStream read(String filePath) {
        try {
            boolean flag = false;
            try {
                //判断sftp上文件是否存在
                sftp.stat(filePath);
                flag = true;
            } catch (SftpException e1) {
                log.error("读取失败: 文件不存在|" + filePath);
            }
            if (flag) {
            	return new ByteArrayInputStream(IoUtil.readBytes(sftp.get(filePath)));
            }
        } catch (Exception e) {
            log.error("读取文件失败:" + filePath, e);
        }
        return null;

    }


    /**
     * 删除文件
     *
     * @param directory  要删除文件所在目录
     * @param deleteFile 要删除的文件
     */
    public void delete(String directory, String deleteFile) throws SftpException {
        sftp.cd(directory);
        sftp.rm(deleteFile);
    }



}